from __future__ import unicode_literals
from __future__ import print_function

from ..elements.utils import attr_bool
from ..elements.boundelement import BoundElement
from ..timezone import Timezone
from ..compat import text_type
from ..context.expressiontime import ExpressionDateTime, ExpressionDate
from ..generickey import GenericKey

from sqlalchemy import (Column,
                        Boolean,
                        ForeignKey,
                        BigInteger,
                        Integer,
                        SmallInteger,
                        String,
                        Float,
                        UnicodeText,
                        DateTime,
                        Sequence)
from sqlalchemy.orm import relationship, backref
from sqlalchemy.ext.mutable import Mutable

from sqlalchemy.types import TypeDecorator, TEXT, DATETIME, DATE
import json
from datetime import datetime, date


class JSONDict(dict):
    def __init__(self, data, json):
        self.update(data)
        self.json = json
        super(JSONDict, self).__init__()


class MoyaCustomDateTime(TypeDecorator):
    """Massage to Moya datetime"""
    impl = DATETIME

    def process_result_value(self, value, dialect):
        if isinstance(value, datetime):
            return ExpressionDateTime.from_datetime(value)
        return value


class MoyaCustomDate(TypeDecorator):
    """Massage to Moya date"""
    impl = DATE

    def process_result_value(self, value, dialect):
        if isinstance(value, date):
            return ExpressionDate.from_date(value)
        return value


class JSONEncodedDict(TypeDecorator):
    "Represents an immutable structure as a json-encoded string."
    impl = TEXT

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


class GenericKeyObject(TypeDecorator):

    impl = TEXT
    def process_bind_param(self, value, dialect):
        if value is not None:
            any_key = GenericKey.from_object(value)
            value = any_key.encode()
        return value

    def process_result_value(self, value, dialect):
        value = GenericKey.decode(value)
        return value


class StringMap(Mutable, dict):

    @classmethod
    def coerce(cls, key, value):
        "Convert plain dictionaries to MutableDict."

        if not isinstance(value, StringMap):
            if isinstance(value, dict):
                return StringMap(value)

            # this call will raise ValueError
            return Mutable.coerce(key, value)
        else:
            return value

    def __setitem__(self, key, value):
        "Detect dictionary set events and emit change events."

        dict.__setitem__(self, key, value)
        self.changed()

    def __delitem__(self, key):
        "Detect dictionary del events and emit change events."

        dict.__delitem__(self, key)
        self.changed()

    def update(self, map):
        dict.update(map)
        self.changed()


#MutableType.associate_with(PickleType)


# class MoyaPickleType(PickleType, Mutable):

#     def result_processor(self, dialect, coltype):
#         impl_processor = self.impl.result_processor(dialect, coltype)
#         loads = self.pickler.loads
#         if impl_processor:
#             def process(value):
#                 value = impl_processor(value)
#                 if value is None:
#                     return None
#                 return MutableObject(loads(value))
#         else:
#             def process(value):
#                 if value is None:
#                     return None
#                 return MutableObject(loads(value))
#         return process


class no_default(object):
    """Sentinal object to signify that no default is required for a column"""
    def __repr__(self):
        return "<no default>"


class MoyaDBColumn(object):
    """Contains information used to create an sqlalchemy column, and its associated abstraction"""
    def __init__(self,
                 type,
                 name,
                 default=no_default,
                 null=True,
                 blank=True,
                 primary=False,
                 index=False,
                 unique=False,
                 label=None,
                 help=None,
                 formfield=None,
                 markup=None):
        self.type = type
        self.name = name
        self.null = null
        self.blank = blank
        if default is not no_default:
            self.default = self.adapt(default)
        else:
            self.default = default
        self.primary = primary
        self.index = index
        self.unique = unique
        self.label = label
        self.help = help
        self.formfield = formfield
        self.markup = markup

    def __repr__(self):
        return "<column {} {}>".format(self.type, self.name)

    def get_dbname(self):
        return self.name

    @property
    def dbname(self):
        return self.get_dbname()

    def get_sa_columns(self):
        kwargs = dict(primary_key=self.primary,
                      nullable=self.null,
                      index=self.index,
                      unique=self.unique)
        if self.default is not no_default:
            kwargs["default"] = self.default
        yield Column(self.name,
                     self.get_sa_type(),
                     **kwargs)

    def get_properties(self, db, table_class):
        return []

    def get_sa_type(self):
        return self.dbtype()

    def adapt(self, value):
        return value

    def get_join(self, node):
        return getattr(node, self.name), None


class PKColumn(MoyaDBColumn):
    """Primary key column"""
    dbtype = Integer

    def adapt(self, value):
        return int(value)

    def get_sa_columns(self):
        sequence_name = "%s_id_seq" % self.name
        kwargs = dict(primary_key=self.primary,
                      nullable=self.null)
        if self.default is not no_default:
            kwargs["default"] = self.default
        yield Column(self.name,
                     self.get_sa_type(),
                     Sequence(sequence_name),
                     **kwargs)


class ForeignKeyColumn(MoyaDBColumn):

    def __init__(self,
                 type,
                 name,
                 ref_model,
                 ref_model_app,
                 label=None,
                 help=None,
                 default=no_default,
                 null=True,
                 blank=True,
                 primary=False,
                 index=False,
                 unique=False,
                 ondelete="CASCADE",
                 orderby=None,
                 options=None,
                 backref=None,
                 uselist=True,
                 picker=None
                 ):
        self.type = type
        self.label = label
        self.help = help
        self.ref_model = BoundElement(ref_model_app, ref_model)
        self.ref_table_name = ref_model.get_table_name(ref_model_app)
        self.name = name
        self.null = null
        self.blank = blank
        if default is not no_default:
            self.default = self.adapt(default)
        else:
            self.default = default
        self.primary = primary
        self.index = index
        self.unique = unique
        self.ondelete = "CASCADE"
        self.orderby = orderby
        self.options = options
        self.backref = backref
        self.uselist = uselist
        self.picker = picker

    def get_dbname(self):
        return self.name + '_id'

    def get_sa_columns(self):
        kwargs = dict(primary_key=self.primary,
                      nullable=self.null,
                      index=self.index,
                      unique=self.unique)
        if self.default is not no_default:
            kwargs["default"] = self.default
        name = "%s_id" % self.name
        yield Column(name,
                     Integer,
                     ForeignKey("%s.id" % self.ref_table_name, ondelete=self.ondelete),
                     **kwargs)

    def get_properties(self, model, table_class):
        "Address.id==Customer.billing_address_id"
        #join = "%s.id==%s.%s_id" % (self.ref_model.name, model.name, self.name)
        def get_join(ref_model_table_class):
            ref_model_table_class = ref_model_table_class()
            join = getattr(table_class, '%s_id' % self.name) == ref_model_table_class.id
            #join = table_class.id == getattr(ref_model_table_class, '%s_id' % self.ref_model.name.lower())
            return join

        def lazy_relationship(app, model):
            if self.backref:
                _backref = backref(self.backref, uselist=self.uselist)
            else:
                _backref = None
            ref_model_table_class = lambda: self.ref_model.element.get_table_class(self.ref_model.app)
            return relationship(ref_model_table_class(),
                                primaryjoin=lambda: get_join(ref_model_table_class),
                                remote_side=lambda: ref_model_table_class().id,
                                backref=_backref,
                                )

        yield self.name, lazy_relationship

    def get_join(self, node):
        ref_model_table_class = self.ref_model.element.get_table_class(self.ref_model.app)
        return ref_model_table_class, getattr(node, self.name)


class BoolColumn(MoyaDBColumn):
    dbtype = Boolean

    def adapt(self, value):
        return attr_bool(value)


class FloatColumn(MoyaDBColumn):
    dbtype = Float

    def adapt(self, value):
        return float(value)


class IntegerColumn(MoyaDBColumn):
    dbtype = Integer

    def adapt(self, value):
        return int(value)


class BigIntegerColumn(IntegerColumn):
    dbtype = BigInteger


class SmallIntegerColumn(IntegerColumn):
    dbtype = SmallInteger


class StringColumn(MoyaDBColumn):
    def __init__(self, type, name, length=None, choices=None, *args, **kwargs):
        self.length = length
        if choices is None:
            self.choices = choices
        else:
            self.choices = list(choices or [])
        super(StringColumn, self).__init__(type, name, *args, **kwargs)

    def get_sa_type(self):
        return String(length=self.length, convert_unicode=True)


class TimezoneType(TypeDecorator):
    '''Prefixes Unicode values with "PREFIX:" on the way in and
    strips it off on the way out.
    '''

    impl = String

    def process_bind_param(self, value, dialect):
        return text_type(value) if value is not None else None

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        try:
            return Timezone(value)
        except:
            return None

    def copy(self):
        return TimezoneType(self.impl.length)


class TimezoneColumn(StringColumn):

    def get_sa_type(self):
        return TimezoneType(length=self.length, convert_unicode=True)


class TextColumn(MoyaDBColumn):
    # def __init__(self, tag_name, name, *args, **kwargs):
    #     super(TextColumn, self).__init__(tag_name, name, *args, **kwargs)

    def get_sa_type(self):
        return UnicodeText()


class DatetimeColumn(MoyaDBColumn):
    dbtype = MoyaCustomDateTime

    def __init__(self, type, name, timezone=False, auto=False, *args, **kwargs):
        self.timezone = timezone
        self.auto = auto
        super(DatetimeColumn, self).__init__(type, name, *args, **kwargs)

    def get_sa_type(self):
        return MoyaCustomDateTime(timezone=self.timezone)


class DateColumn(MoyaDBColumn):
    dbtype = MoyaCustomDate

    def __init__(self, type, name, auto=False, *args, **kwargs):
        self.auto = auto
        super(DateColumn, self).__init__(type, name, *args, **kwargs)

    def get_sa_type(self):
        return MoyaCustomDate()


class StringMapColumn(MoyaDBColumn):
    #dbtype = PickleType

    def get_sa_type(self):
        return StringMap.as_mutable(JSONEncodedDict)


class GenericKeyColumn(MoyaDBColumn):
    dbtype = GenericKeyObject




from __future__ import unicode_literals
from __future__ import print_function

from ..elements.elementbase import Attribute
from ..tags.context import DataSetter
from ..template.rendercontainer import RenderContainer
from ..logic import DeferNodeContents, EndLogic
from ..render import render_object
from ..response import MoyaResponse
from ..compat import PY2, py2bytes


class RenderTemplate(DataSetter):
    """Render a template"""

    class Help:
        synopsis = "render a template and store the result"

    template = Attribute("Template", type="template", required=True)
    withscope = Attribute("Use current scope?", default=False, type="boolean")
    format = Attribute("Format to render", default="html", required=False)
    _from = Attribute("Application", type="application", required=False, default=None)

    def logic(self, context):
        params = self.get_parameters(context)
        app = self.get_app(context)
        template = app.resolve_template(params.template)
        value = RenderContainer.create(app,
                                       template=template)
        value.update(self.get_let_map(context))
        if params.withscope:
            value.update(context['.call'])
        context['_render'] = value
        with context.scope('_render'):
            yield DeferNodeContents(self)
        self.on_value(context, value)

    def on_value(self, context, value):
        html = render_object(value, self.archive, context, self.format(context))
        self.set_context(context, self.dst(context), html)


class ServeTemplate(RenderTemplate):
    """Render and serve a template"""

    content_type = Attribute("Mime Type", required=False, default=None)

    class Help:
        synopsis = """render and serve a template"""

    def on_value(self, context, value):
        content_type = self.content_type(context)
        html = render_object(value, self.archive, context, self.format(context))
        response = MoyaResponse(charset=py2bytes('utf8'))
        if content_type:
            response.content_type = py2bytes(content_type)
        response.text = html
        raise EndLogic(response)



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import

from .compat import implements_to_string
from .html import slugify
from .tools import remove_padding
from .filter import MoyaFilterParams
from . import namespaces

from collections import namedtuple

from fs.path import relativefrom, dirname

import re
import textwrap


class BBCodeError(Exception):
    pass


class MultiReplace(object):
    def __init__(self, repl_dict):
        # string to string mapping; use a regular expression
        keys = sorted(repl_dict.keys(), reverse=True)
        #keys.sort(reverse=True)  # lexical order
        pattern = "|".join([re.escape(key) for key in keys])
        self.pattern = re.compile(pattern)
        self.dict = repl_dict
        self.sub = self.pattern.sub

    def replace(self, s):
        # apply replacement dictionary to string
        get = self.dict.get

        def repl(match):
            item = match.group(0)
            return get(item, item)
        return self.sub(repl, s)

    __call__ = replace


html_escape = MultiReplace({'<': '&lt;',
                            '>': '&gt;',
                            '&': '&amp;',
                            '\n': '<br/>'})


@implements_to_string
class Tag(object):

    inline = True
    enclosed = False
    auto_close = False
    visible = True

    html_open = ""
    html_close = ""

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return "[{}]".format(self.name)

    def match(self, tag_name):
        return tag_name.lower() == self.name

    def on_insert(self, attribs, data):
        self.attribs = attribs

    def render_open(self, data):
        return self.html_open

    def render_close(self, data, text=None):
        return self.html_close


class BlockTag(Tag):
    inline = False


class FieldTag(Tag):
    auto_close = True

    def on_insert(self, attribs, data):
        data[self.name] = attribs


class ParagraphTag(Tag):
    inline = False
    html_open = "<p>"
    html_close = "</p>\n"


class SimpleBlockTag(BlockTag):
    def __init__(self, name, cls):
        super(SimpleBlockTag, self).__init__(name)
        self.html_open = '<div class="{}">'.format(cls)
        self.html_close = '</div>'


class InlineTag(Tag):
    inline = True
    enclosed = False
    auto_close = False
    visible = True

    def __init__(self, name, html_open, html_close):
        super(InlineTag, self).__init__(name)
        self.html_open = html_open
        self.html_close = html_close

    def render_open(self, data):
        return self.html_open

    def render_close(self, data, text=None):
        return self.html_close


class InlineCodeTag(Tag):
    inline = True
    enclosed = True
    auto_close = False
    visible = True

    def render_close(self, data, text=None):
        return '<code>{}</code>'.format(html_escape(text))


class SettingTag(InlineCodeTag):

    def render_close(self, data, text=None):
        if '=' in text:
            k, v = text.split('=', 1)
            k = html_escape(k)
            v = '<span class="value">{}</span>'.format(html_escape(v))
            return '<p class="setting">{}={}</p>'.format(k, v)
        return '<p class="setting">{}</p>'.format(html_escape(text))


class RawTag(Tag):
    inline = True
    visible = True
    enclosed = True

    def render_close(self, data, text=None):
        return text or ''


class AlertTag(BlockTag):
    html_close = "</div>"

    def render_open(self, data):
        html = '<div class="alert alert-warning">'
        if self.attribs.strip():
            html += '<strong>{}</strong> '.format(self.attribs)
        return html


class AsideTag(BlockTag):
    html_close = "</aside>"

    def render_open(self, data):
        html = '<aside>'
        if self.attribs.strip():
            html += '<strong>{}</strong> '.format(self.attribs)
        return html


class NoteTag(BlockTag):
    html_open = '<p class="note">'
    html_close = "</p>"


class CodeTag(BlockTag):
    enclosed = True
    inline = True

    def render_open(self, data):
        self.lang = self.attribs.strip().lower()
        return '''<pre class="moya-console format-{}">'''.format(self.lang)

    def render_close(self, data, text=None):
        from .syntax import highlight
        html = highlight(self.lang, remove_padding(text), line_numbers=False)

        return html + '</pre>'


class H1Tag(BlockTag):
    inline = False
    enclosed = True

    level = 1

    def render_close(self, data, text=None):
        title = text.strip()
        data.setdefault('docmap', []).append([self.level, title])
        h_format = '<h{level}><a name="{anchor}"></a><a href="#{anchor}">{text}<span class="anchor"> &#182;</span></a></h{level}>'
        return h_format.format(text=html_escape(text),
                               level=self.level + 1,
                               anchor=slugify(title))


class H2Tag(H1Tag):
    level = 2


class H3Tag(H1Tag):
    level = 3


class H4Tag(H1Tag):
    level = 4


Index = namedtuple("Index", ['type', 'template', 'lines'])


class IndexTag(BlockTag):
    enclosed = True

    def render_close(self, data, text=None):
        text = text or ''
        index_name = self.attribs.strip() or 'main'
        index_type = '1'
        template = None
        if ' ' in index_name:
            index_name, index_type = index_name.split(' ', 1)
            if ' ' in index_type:
                index_type, template = index_type.split(' ', 1)
        lines = [line.strip() for line in text.splitlines() if line.strip()]

        data.setdefault('indices', {})[index_name] = Index(index_type, template, lines)
        return "{{{" + "INDEX {}".format(index_name) + "}}}"


class URLTag(Tag):
    inline = True

    def render_open(self, data):
        url = self.attribs.strip()
        return '''<a href="{}">'''.format(url)

    def render_close(self, data, text=None):
        return "</a>"


class DocTag(Tag):

    inline = True
    auto_close = True

    _format = '''<a href="{url}">{title}</a>'''

    def render_open(self, data):
        pass

    def render_close(self, data, text=None):
        doc_name = self.attribs.strip()
        fragment = None
        if '#' in doc_name:
            doc_name, fragment = doc_name.split('#', 1)
        if 'docs' not in data:
            return doc_name
        doc = data['docs'].get('doc.{}'.format(doc_name))
        if not doc:
            return doc_name
        title = doc.data.get('title', doc_name)
        urls = data['urls']
        url = urls['doc'].get(doc_name, '')
        if fragment is not None:
            url = url + '#' + fragment
        context = data['context']
        path = dirname(context.get('.request.path', '/'))
        return self._format.format(title=title, url=relativefrom(path, url))


class DocLinkTag(Tag):
    inline = True
    auto_close = False

    def render_open(self, data):
        doc_name = self.attribs.strip()
        anchor = None
        if '#' in doc_name:
            doc_name, anchor = doc_name.split('#', 1)
        if 'docs' not in data:
            return doc_name
        doc = data['docs'].get('doc.{}'.format(doc_name))
        if not doc:
            return doc_name
        title = doc.data.get('title', doc_name)
        urls = data['urls']
        url = urls['doc'].get(doc_name, '')
        context = data['context']
        path = dirname(context.get('.request.path', '/'))
        if anchor:
            url = "{}#{}".format(url, anchor)
        return '''<a href="{url}" title="{title}">'''.format(url=relativefrom(path, url), title=title)

    def render_close(self, data, text=None):
        return "</a>"


class TagTag(BlockTag):
    inline = True
    auto_close = False
    enclosed = True

    _re_namespace = re.compile(r"^\{(.*?)\}(.*?)$")

    @classmethod
    def _join(cls, ns, name):
        if name:
            return "{}/{}".format(ns, name)
        return ns

    def render_open(self, data):
        return ''

    def render_close(self, data, text=None):
        if 'context' not in data:
            return "<code>{}</code>".format(text)
        context = data['context']
        path = dirname(context.get('.request.path', '/'))
        urls = context['.urls']
        tag_name = text.strip()
        xmlns = None
        if '{' in tag_name:
            xmlns, tag_name = self._re_namespace.match(tag_name).groups()
            if '://' not in xmlns:
                xmlns = self._join(namespaces.default, xmlns)
            text = tag_name
            tag_name = "{{{}}}{}".format(xmlns, tag_name)

        if xmlns is None:
            xmlns = self.attribs.strip()
            if '://' not in xmlns:
                xmlns = self._join(namespaces.default, xmlns)
            tag_name = "{{{}}}{}".format(xmlns, tag_name)
        try:
            tag_path = urls['tag'][tag_name]
            relative_tag_path = relativefrom(path, tag_path)
        except KeyError as e:
            return "<code>{}</code>".format(text)
        else:
            return '''<a class="tag" href="{tag_path}">&lt;{text}&gt;</a>'''.format(tag_path=relative_tag_path, text=text)


class DefinitionsTag(BlockTag):

    def render_open(self, data):
        return '<dl class="dl-horizontal">'

    def render_close(self, data, text=None):
        return "</dl>"


class DefineTag(BlockTag):
    inline = True

    def render_open(self, data):
        return '<dt>{}</dt>\n<dd>'.format(html_escape(self.attribs))

    def render_close(self, data, text=None):
        return '</dd>'


class BreakTag(BlockTag):
    inline = True
    auto_close = True

    def render_close(self, data, text=None):
        return "<br>"



_re_remove_markup = re.compile(r'\[.*?\]', re.DOTALL | re.UNICODE)
_re_break_groups = re.compile('[\n]{2,}', re.DOTALL | re.UNICODE)


class BBCode(object):

    standard_replace = MultiReplace({'<': '&lt;',
                                     '>': '&gt;',
                                     '&': '&amp;',
                                     '\n': '<br/>'})

    standard_unreplace = MultiReplace({'&lt;': '<',
                                       '&gt;': '>',
                                       '&amp;': '&'})

    standard_replace_no_break = MultiReplace({'<': '&lt;',
                                              '>': '&gt;',
                                              '&': '&amp;'})

    cosmetic_replace = MultiReplace({'--': '&ndash;',
                                     '---': '&mdash;',
                                     '...': '&#8230;',
                                     '(c)': '&copy;',
                                     '(reg)': '&reg;',
                                     '(tm)': '&trade;'})

    _re_tag_on_line = re.compile(r'\[.*?\]', re.UNICODE)
    _re_end_eq = re.compile(r"\]|\=", re.UNICODE)
    _re_quote_end = re.compile(r'\"|\]', re.UNICODE)
    _re_tag_token = re.compile(r'^\[(\S*?)[\s=]\"?(.*?)\"?\]$', re.UNICODE)
    _re_new_paragraph = re.compile(r'\n*?', re.UNICODE)

    def __init__(self):
        self.registry = []
        self.data = {}
        self.add_tag(ParagraphTag, 'p')

    def add_tag(self, tag_class, *args, **kwargs):
        tag_instance = tag_class(*args, **kwargs)
        self.registry.append(tag_instance)

    @classmethod
    def get_locations(cls, post):
        pos_to_location = {}
        line_start = 0
        for line_no, line in enumerate(post.splitlines(True)):
            line_length = len(line)
            for row in range(line_length):
                pos_to_location[line_start + row] = (line_no, row)
            line_start += line_length
        return pos_to_location

    @classmethod
    def tokenize(cls, post):
        locations = cls.get_locations(post)
        re_tag_on_line = cls._re_tag_on_line
        re_end_eq = cls._re_end_eq
        re_quote_end = cls._re_quote_end
        pos = 0

        def find_first(post, pos, re_ff):
            search = re_ff.search(post, pos)
            if search is None:
                return -1
            return search.start()

        TOKEN_TAG, TOKEN_PTAG, TOKEN_TEXT, TOKEN_PARAGRAPH = range(4)

        def yield_text(text):
            while '\n\n' in text:
                old_paragraph, text = text.split('\n\n', 1)
                if old_paragraph:
                    yield TOKEN_TEXT, old_paragraph
                yield TOKEN_PARAGRAPH, '\n\n'
            if text:
                yield TOKEN_TEXT, text

        post_find = post.find
        while True:
            brace_pos = find_first(post, pos, re_tag_on_line)
            if brace_pos == -1:
                if pos < len(post):
                    for tag_type, text in yield_text(post[pos:]):
                        yield locations[pos], tag_type, text
                    #yield TOKEN_TEXT, post[pos:]
                return
            if brace_pos - pos > 0:
                for tag_type, text in yield_text(post[pos:brace_pos]):
                    yield locations[pos], tag_type, text
                #yield TOKEN_TEXT, post[pos:brace_pos]

            pos = brace_pos
            end_pos = pos + 1

            open_tag_pos = post_find('[', end_pos)
            end_pos = find_first(post, end_pos, re_end_eq)
            if end_pos == -1:
                for tag_type, text in yield_text(post[pos:]):
                    yield locations[pos], tag_type, text
                #yield TOKEN_TEXT, post[pos:]
                return

            if open_tag_pos != -1 and open_tag_pos < end_pos:
                for tag_type, text in yield_text(post[pos:open_tag_pos]):
                    yield locations[pos], tag_type, text
                #yield TOKEN_TEXT, post[pos:open_tag_pos]
                end_pos = open_tag_pos
                pos = end_pos
                continue

            if post[end_pos] == ']':
                yield locations[pos], TOKEN_TAG, post[pos:end_pos + 1]
                pos = end_pos + 1
                continue

            if post[end_pos] == '=':
                try:
                    end_pos += 1
                    while post[end_pos] == ' ':
                        end_pos += 1
                    if post[end_pos] != '"':
                        end_pos = post_find(']', end_pos + 1)
                        if end_pos == -1:
                            return
                        for tag_type, text in yield_text(post[pos:end_pos + 1]):
                            yield locations[pos], tag_type, text
                    else:
                        end_pos = find_first(post, end_pos, re_quote_end)
                        if end_pos == -1:
                            return
                        if post[end_pos] == '"':
                            end_pos = post_find('"', end_pos + 1)
                            if end_pos == -1:
                                return
                            end_pos = post_find(']', end_pos + 1)
                            if end_pos == -1:
                                return
                            yield locations[pos], TOKEN_PTAG, post[pos:end_pos + 1]
                        else:
                            yield locations[pos], TOKEN_TAG, post[pos:end_pos + 1]
                    pos = end_pos + 1
                except IndexError:
                    return

    @classmethod
    def parse_tag_token(cls, s):
        m = cls._re_tag_token.match(s.lstrip())
        if m is None:
            name, attribs = s[1:-1], u''
        else:
            name, attribs = m.groups()
        if name.startswith(u'/'):
            return name.strip()[1:].lower(), attribs.strip(), True
        else:
            return name.strip().lower(), attribs.strip(), False

    # Matches simple blank tags containing only whitespace
    _re_blank_tags = re.compile(r"\<(\w+?)\>\</\1\>")
    _re_blank_with_spaces_tags = re.compile(r"\<(\w+?)\>\s+\</\1\>")
    _re_whitespace_word = re.compile(r"(\s+?\S*)")

    @classmethod
    def cleanup_html(cls, html):
        """Cleans up html. Currently only removes blank tags, i.e. tags containing only
        whitespace. Only applies to tags without attributes. Tag removal is done
        recursively until there are no more blank tags. So <strong><em></em></strong>
        would be completely removed.

        html -- A string containing (X)HTML

        """
        original_html = ''
        while original_html != html:
            original_html = html
            html = cls._re_blank_tags.sub(" ", html)
            html = cls._re_blank_with_spaces_tags.sub(" ", html)
        #html = _re_break_groups.sub("\n", html)
        return html

    def wrap(self, bbcode, max_length=79):
        TOKEN_TAG, TOKEN_PTAG, TOKEN_TEXT, TOKEN_PARAGRAPH = range(4)
        lines = []
        line_length = 0

        bbcode = textwrap.dedent(bbcode)
        for l in bbcode.splitlines():
            lines.append([])
            line_length = 0
            for loc, tag_type, tag_token in self.tokenize(l):
                if not lines[-1]:
                    tag_token = tag_token.lstrip()
                if not tag_token:
                    continue
                if tag_type == TOKEN_TEXT:
                    for word in self._re_whitespace_word.split(tag_token):
                        if not word:
                            continue
                        if line_length + len(word) > max_length:
                            word = word.lstrip()
                            lines.append([word])
                            line_length = len(word)
                        else:
                            lines[-1].append(word)
                            line_length += len(word)
                else:
                    lines[-1].append(tag_token)

        return '\n\n'.join(''.join(word for word in line) for line in lines)

    def render_console(self, bbcode, max_length=79):
        from .console import AttrText, style, XMLHighlighter
        bbcode = self.wrap(bbcode.strip(), max_length=min(120, max_length)) + '\n'
        TOKEN_TAG, TOKEN_PTAG, TOKEN_TEXT, TOKEN_PARAGRAPH = range(4)
        _bbcode_map = {
            'b': 'bold',
            'i': 'italic',
            'c': 'bold cyan',
            'u': 'underline',
            'd': 'dim',
            'tag': 'bold cyan',
            'code': 'bold',
            'note': 'italic',
            'error': 'bold red',
            'success': 'bold green'
        }
        tag_stack = []
        text = []
        pos = 0
        attributes = []
        for loc, tag_type, tag_token in self.tokenize(bbcode):
            if tag_type == TOKEN_TEXT:
                text += tag_token
                pos += len(tag_token)
            elif tag_type in (TOKEN_TAG, TOKEN_PTAG):
                tag_name, tag_attribs, end_tag = self.parse_tag_token(tag_token)
                if end_tag:
                    try:
                        if tag_name == tag_stack[-1][0]:
                            tag_name, attribute_start, tag_style = tag_stack.pop()
                            attributes.append((attribute_start, pos, tag_style))
                    except IndexError:
                        raise ValueError("end tag {} doesn't match an opening tag".format(tag_token))
                else:
                    if tag_name in _bbcode_map:
                        tag_style = style(_bbcode_map[tag_name])
                        tag_stack.append((tag_name, pos, tag_style))
                    else:
                        tag_stack.append((tag_name, pos, None))

            elif tag_type == TOKEN_PARAGRAPH:
                text.append('\n')
                pos += 1

        text = AttrText(''.join(text))
        text = XMLHighlighter.highlight(text)
        for start, end, tag_style in attributes:
            if tag_style is not None:
                text.add_span(start=start, end=end, **tag_style)
        return text

    def render(self, text, data=None, path="?"):
        TOKEN_TAG, TOKEN_PTAG, TOKEN_TEXT, TOKEN_PARAGRAPH = range(4)

        if data is None:
            data = {}

        def raise_error(text):
            raise BBCodeError('File "{}", line {}, col {}: {}'.format(path, line + 1, col + 1, text))

        html = []
        add_html = html.append
        tag_stack = []
        enclosed_text = []
        add_enclosed_text = enclosed_text.append

        def html_escape(s):
            return s.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')

        for (line, col), tag_type, tag_token in self.tokenize(text):
            if tag_stack and tag_stack[-1].enclosed:
                if tag_type in (TOKEN_TAG, TOKEN_PTAG):
                    tag_name, tag_attribs, end_tag = self.parse_tag_token(tag_token)
                    if end_tag and tag_name.lower() == tag_stack[-1].name:
                        try:
                            tag = tag_stack.pop()
                        except IndexError:
                            raise_error("Unexpected close tag: {}".format(tag_token))
                        text = ''.join(enclosed_text)
                        add_html(tag.render_close(data, text=text) or '')
                        del enclosed_text[:]
                    else:
                        add_enclosed_text(tag_token)
                else:
                    add_enclosed_text(tag_token)
                continue

            if tag_type == TOKEN_TEXT:
                if not tag_stack:
                    ptag = self.get_tag("p", '', data)
                    tag_stack.append(ptag)
                    add_html(ptag.render_open(data) or '')
                if not enclosed_text:
                    add_html(self.cosmetic_replace(html_escape(tag_token)))
                else:
                    add_html(html_escape(tag_token))

            elif tag_type == TOKEN_PARAGRAPH:
                if tag_stack and tag_stack[-1].name == "p":
                    tag = tag_stack.pop()
                    add_html(tag.render_close(data) or '')

            elif tag_type in (TOKEN_TAG, TOKEN_PTAG):
                tag_name, tag_attribs, end_tag = self.parse_tag_token(tag_token)

                if end_tag:
                    try:
                        tag = tag_stack.pop()
                    except IndexError:
                        raise_error("unexpected close tag '{}'".format(tag_token))
                    if tag_name.lower() != tag.name:
                        raise_error("mismatched close tag '{}'".format(tag_name))
                    add_html(tag.render_close(data) or '')

                else:
                    tag = self.get_tag(tag_name, tag_attribs, data)
                    if tag is None:
                        raise_error("unknown tag '{}'".format(tag_name))

                    if not tag.inline:
                        while tag_stack:
                            add_html(tag_stack.pop().render_close(data) or '')

                    tag_stack.append(tag)
                    add_html(tag.render_open(data) or '')
                    if tag.auto_close:
                        tag_stack.pop()
                        add_html(tag.render_close(data) or '')

        while tag_stack:
            tag = tag_stack.pop()
            add_html(tag.render_close(data, text='') or '')

        return self.cleanup_html(''.join(html)).strip(), data

    def __call__(self, text):
        html, data = self.render(text)
        return html

    def __repr__(self):
        return "<bbcode parser>"

    def __moyafilter__(self, context, app, value, **params):
        data = {'context': context,
                'docs': context.get('.docs', {}),
                'urls': context.get('.urls', {})}
        html, data = self.render(value, data=data, path=params.get('path', 'unknown'))
        return html

    def __moyacall__(self, params):
        return MoyaFilterParams(self, params)

    def get_tag(self, tag_name, attribs, data):
        for tag_instance in self.registry:
            if tag_instance.match(tag_name):
                tag_instance.on_insert(attribs, data)
                return tag_instance
        return None


parser = BBCode()
add_tag = parser.add_tag
add_tag(InlineTag, 'i', '<em>', '</em>')
add_tag(InlineTag, 'b', '<b>', '</b>')
add_tag(InlineTag, 'u', '<u>', '</u>')
add_tag(InlineTag, 's', '<s>', '</s>')
add_tag(TagTag, 'tag')
add_tag(SettingTag, 'setting')
add_tag(InlineCodeTag, 'c')
add_tag(H1Tag, 'h1')
add_tag(H2Tag, 'h2')
add_tag(H3Tag, 'h3')
add_tag(H4Tag, 'h4')
add_tag(URLTag, 'url')
add_tag(RawTag, 'raw')

add_tag(AlertTag, 'alert')
add_tag(AsideTag, 'aside')
add_tag(NoteTag, 'note')

add_tag(IndexTag, 'index')
add_tag(IndexTag, 'appendix')
add_tag(FieldTag, 'title')
add_tag(FieldTag, 'class')
add_tag(FieldTag, 'section')
add_tag(FieldTag, 'name')
add_tag(FieldTag, 'id')
add_tag(CodeTag, 'code')
add_tag(DocTag, 'doc')
add_tag(DocLinkTag, 'link')
add_tag(DefinitionsTag, 'definitions')
add_tag(DefineTag, 'define')
add_tag(BreakTag, 'br')

render = parser
render_console = parser.render_console


if __name__ == "__main__":
    text = """This is a [i]test[/i] of [b]bbcode[/b] rendering in the console.

    <echo>${hobbits}</echo>

    """
    text = "[h1]te]bst[/c]"
    render(text)
    # from moya.console import Console
    # c = Console()
    # parser.render_console(text)

    # bbcode = "This is long [b]bbcode[/b] that should be wrapped on to multiple lines. Let's see if it [i]works![/i]"
    # print(parser.wrap(bbcode, 30))



from __future__ import unicode_literals

from ...command import SubCommand
from ...wsgi import WSGIApplication
from ...console import Cell


class Settings(SubCommand):
    """Manage moya project settings"""
    help = "manage settings"

    def add_arguments(self, parser):
        parser.add_argument(dest="name", default=None, metavar="NAME", nargs='?',
                            help="display settings for a specific library or application")
        parser.add_argument("-l", "--location", dest="location", default=None, metavar="PATH",
                            help="location of the Moya server code")
        parser.add_argument("-i", "--ini", dest="settings", default=None, metavar="SETTINGSPATH",
                            help="path to projects settings file")
        parser.add_argument("--server", dest="server", default='main', metavar="SERVERREF",
                            help="server element to use")
        return parser

    def run(self):
        args = self.args
        application = WSGIApplication(self.location, self.get_settings(), args.server, disable_autoreload=True)
        archive = application.archive

        libstyle = dict(bold=True, fg="magenta")
        appstyle = dict(bold=True, fg="green")

        if args.name is not None:
            try:
                if '.' in args.name:
                    libs = [(args.name, archive.libs[args.name])]
                    apps = []
                else:
                    apps = [(args.name, archive.apps[args.name])]
                    libs = []
            except KeyError:
                self.console.error("'%s' is not an application or library in this project" % args.name)
                return -1
        else:
            libs = archive.libs.items()
            apps = archive.apps.items()

        for name, lib in sorted(libs):
            if lib.settings:
                self.console.nl()("[LIB] %s settings" % name, **libstyle).nl()
                table = []
                table.append([Cell("Setting", bold=True), Cell("Value", bold=True)])
                for k, v in lib.settings.items():
                    table.append([Cell(k), Cell(v)])
                self.console.table(table)

        for name, app in sorted(apps):
            if app.settings:
                self.console.nl()("[APP] %s: %s settings" % (app.lib.long_name, app.name), **appstyle).nl()

                table = []
                table.append([Cell("Setting", bold=True), Cell("Value", bold=True)])
                for k, v in app.settings.items():
                    table.append([Cell(k), Cell(v)])
                self.console.table(table)



from __future__ import unicode_literals

__author__ = "Will McGugan <admin@moyaproject.com>"
__version__ = "0.5.12"
# *** Don't forget to update version in setup.py ***

__all__ = ['pilot', 'expose']


from . import expose
from .moyapilot import Pilot
from .compat import text_type


pilot = Pilot()

import decimal


class MoyaDecimal(decimal.Decimal):

    def __moyarepr__(self, context):
        return "d:'{}'".format(text_type(self))


decimal.Decimal = MoyaDecimal



from __future__ import unicode_literals

from ..elements.elementbase import LogicElement, Attribute
from ..logic import DeferNodeContents

from time import time, clock

import logging
log = logging.getLogger('moya.runtime')


class Timer(LogicElement):
    """
    Time a block of code, and log the elapsed time. The log will display two time values; the first is the wall time, the second is [i]processor time[/i]. Processor time is more useful for profiling, as it doesn't include the time spent waiting on IO.

    """

    msg = Attribute("msg", default="elapsed")
    ms = Attribute("Display milliseconds", type="boolean", default=False)
    console = Attribute("Display to console?", type="boolean", default=False)

    class Help:
        synopsis = "time a block of code"
        help = """
        <timer msg="time slow macro">
            <call macro="#slow" />
        </timer>
        """

    def logic(self, context):
        to_console = self.console(context)
        msg = self.msg(context)
        ms = self.ms(context)
        start = time()
        start_clock = clock()
        try:
            yield DeferNodeContents(self)
        finally:
            taken_clock = clock() - start_clock
            taken = time() - start
            if ms:
                result = ("{}: {:.2f}ms {:.2f}ms".format(msg, taken * 1000, taken_clock * 1000))
            else:
                result = ("{}: {:.2f}s {:.2f}s".format(msg, taken, taken_clock))
            if to_console:
                context['.console'].text(result)
            else:
                log.debug(result)


class ProfilePython(LogicElement):
    """
    Profile Python code.

    """

    class Help:
        synopsis = "profile python code"

    _sort_choices = [
        'calls',  # call count
        'cumulative',  #  cumulative time
        'cumtime',  # cumulative time
        'file',  # file name
        'filename',  # file name
        'module',  #    file name
        'ncalls',  #    call count
        'pcalls',  #    primitive call count
        'line',  #  line number
        'name',  #  function name
        'nfl',  #   name/file/line
        'stdname',  #   standard name
        'time',  #  internal time
        'tottime'
    ]
    sort = Attribute("column to sort by", choices=_sort_choices, default="cumtime")
    max = Attribute("number of stats to display", type="integer", default=20)

    def logic(self, context):
        sort, max_stats = self.get_parameters(context, 'sort', 'max')
        import cProfile, pstats
        prof = cProfile.Profile()
        prof.enable()
        try:
            yield DeferNodeContents(self)
        finally:
            prof.disable()
            stats = pstats.Stats(prof)
            stats.sort_stats(sort)
            stats.print_stats(max_stats)



from __future__ import unicode_literals
from __future__ import print_function

from moya.wsgi import WSGIApplication
from moya.sites import Sites
from moya.settings import SettingsContainer
from moya.compat import py2bytes, itervalues

from webob import Response

import sys
import os
import io
import glob
from collections import OrderedDict


DEFAULT_HOME_DIR = "/etc/moya"

not_found_response = """<!DOCTYPE html>
<html>
<head>
    <title>404 Not Found</title>
    <style type="text/css">
        body {{font-family: arial,sans-serif;}}
    </style>
</head>
<body>
<h1>404 Not Found</h1>
</body>
</html>

"""


class Server(object):
    def __init__(self, name, domains, location, ini, logging, master_settings=None, master_logging=None):
        self.name = name
        self.domains = domains
        self.location = location
        self.ini = ini
        self.logging = logging
        self.master_logging = master_logging
        self.master_settings = master_settings

        self.application = None

    def build(self):
        application = WSGIApplication(self.location,
                                      self.ini,
                                      logging=self.logging,
                                      master_settings=self.master_settings,
                                      master_logging=self.master_logging)
        application.build()
        self.application = application


class MultiWSGIApplication(object):

    def __init__(self):
        self.servers = OrderedDict()
        self.sites = Sites()

    def add_project(self, settings):
        name = settings.get('service', 'name')
        domains = settings.get_list('service', 'domains')
        location = os.path.join(self.home_dir, settings.get('service', 'location'))
        ini = settings.get_list('service', 'ini') or ['production.ini']
        logging_setting = settings.get('service', 'logging', None)
        if logging_setting is None:
            logging = 'logging.ini'
            master_logging = None
        else:
            logging = None
            master_logging = os.path.join(self.home_dir, logging_setting)

        server = Server(name, domains, location, ini, logging, master_settings=settings, master_logging=master_logging)
        self.servers[name] = server
        self.sites.add(domains, name=name)

    def build_all(self):
        for server in itervalues(self.servers):
            server.build()

    def not_found(self):
        response = Response(charset=py2bytes("utf8"), status=404)
        response.text = not_found_response

    def __call__(self, environ, start_response):
        domain = environ['SERVER_NAME']
        site_match = self.sites.match(domain)
        if site_match is None:
            return self.not_found()
        server_name = site_match.data['name']
        server_application = self.servers[server_name].application
        return server_application(environ, start_response)


class Service(MultiWSGIApplication):

    def error(self, msg, code=-1):
        sys.stderr.write(msg + '\n')
        sys.exit(code)

    def __init__(self, home_dir=None):
        super(Service, self).__init__()
        self.home_dir = home_dir = os.environ.get('MOYA_SRV_HOME', None) or DEFAULT_HOME_DIR

        settings_path = os.path.join(home_dir, 'moya.conf')
        try:
            with io.open(settings_path, 'rt') as f:
                self.settings = SettingsContainer.read_from_file(f)
        except IOError:
            self.error('unable to read {}'.format(settings_path))
            return -1

        for path in self._get_projects():
            settings = self._read_project(path)
            self.add_project(settings)

        self.build_all()

    def _get_projects(self):
        project_paths = self.settings.get_list('projects', 'read')
        paths = []
        cwd = os.getcwd()
        try:
            os.chdir(self.home_dir)
            for path in project_paths:
                glob_paths = glob.glob(path)
                paths.extend([os.path.abspath(p) for p in glob_paths])
        finally:
            os.chdir(cwd)

        return paths

    def _read_project(self, path):
        settings = SettingsContainer.read_os(path)
        return settings



from __future__ import unicode_literals
from __future__ import print_function

from ..console import make_table_header
from ..tools import extract_namespace
from .. import namespaces
from ..console import Cell
from .. import bbcode

import textwrap


def help(archive, console, tagname):
    """Generate console help"""
    xmlns, tagname = extract_namespace(tagname)
    if '/' not in xmlns:
        xmlns = namespaces.default + '/' + xmlns
    tag = archive.registry.get_tag("{{{}}}{}".format(xmlns, tagname))
    if tag is None:
        if xmlns and not archive.registry.check_namespace(xmlns):
            console.error("No such namespace: %s" % xmlns)
        else:
            console.error("No such tag: %s" % tagname)
        return False
    tag_name = tag._tag_name
    doc = "<%s/>" % tag_name

    doc_attribs = []

    for name, attrib in sorted(tag._tag_attributes.items()):
        if attrib.metavar:
            metavar = attrib.metavar.upper()
            doc_attribs.append('%s="%s"' % (attrib.name or name, metavar))
    if doc_attribs:
        doc = "<%s %s />" % (tag_name, (" ".join(doc_attribs)))
    else:
        doc = "<%s/>" % tag_name

    console.xml(doc.strip()).nl()

    if tag._tag_doc:
        console(bbcode.render_console(tag._tag_doc, max_length=console.width)).nl()

    details = [(Cell("name", bold=True), tagname),
               (Cell("namespace", bold=True), xmlns),
               (Cell("defined", bold=True), getattr(tag, '_definition', '?'))
               ]
    console.table(details, header=False, dividers=False, grid=False).nl()

    if hasattr(tag, 'Help'):
        example = getattr(tag.Help, 'example', None)
        if example:
            example = textwrap.dedent(example)
            console('[example(s)]', fg="magenta", bold=True).nl()
            console.xml(example).nl()

    base_attributes = tag._get_base_attributes()
    params = []
    for name, attrib in tag._tag_attributes.items():
        if name not in base_attributes:
            params.append((Cell('"{}"'.format(name), fg="cyan"),
                          attrib.doc,
                          attrib.type_display.lower(),
                          'Yes' if attrib.required else 'No',
                           attrib.default_display(attrib.default) if not attrib.required else ''
                           ))
    params.sort(key=lambda p: p[0])
    if params:
        console("[attributes]", fg="green", bold=True).nl()
        console.table(make_table_header("attrib", "doc", "type", "required?", "default") + params).nl()

    params = []
    for name, attrib in base_attributes.items():
        if name in tag._tag_attributes:
            docattrib = tag._tag_attributes[name]
        else:
            docattrib = attrib
        params.append((Cell('"{}"'.format(name), fg="cyan"),
                      docattrib.doc,
                      docattrib.type_display.lower(),
                      'Yes' if docattrib.required else 'No',
                       docattrib.default if not docattrib.required else ''
                       ))
    if params:
        console("[inherited attributes]", fg="green", bold=True).nl()
        console.table(make_table_header("attrib", "doc", "type", "required?", "default") + params)

    return True



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import

from sqlalchemy import create_engine, MetaData
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm import scoped_session
from sqlalchemy.exc import DatabaseError, IntegrityError, OperationalError, StatementError

from . import namespaces
from .elements.utils import attr_bool
from .compat import text_type, implements_to_string, itervalues
from . import logic

import logging
startup_log = logging.getLogger('moya.startup')
db_log = logging.getLogger('moya.db')


def dbobject(obj):
    return getattr(obj, '__moyadbobject__', lambda: obj)()


@implements_to_string
class DBEngine(object):
    def __init__(self, name, engine_name, engine, default=False):
        self.name = name
        self.engine_name = engine_name
        self.engine = engine
        self.default = default
        self.Session = sessionmaker(bind=engine)
        self.metadata = MetaData()

    def get_session(self):
        return DBSession(self.Session, self.engine)

    def __str__(self):
        return '<dbengine %s>' % self.engine_name

    def __repr__(self):
        return '<dbengine "%s">' % self.name


def _get_db_error(e):
    """Extract information from sqlalchemy error"""
    message = getattr(e, 'message', text_type(e))
    info = {'sql': e.statement, 'params': e.params}
    if hasattr(e, 'orig'):
        try:
            code, message = e.orig.args
        except:
            pass
        else:
            info['code'] = code
            message = message
    return message, info


def wrap_db_errors(f):
    """Turn DB errors in to moya errors"""
    def deco(self, *args, **kwargs):
        try:
            ret = f(self, *args, **kwargs)
        except IntegrityError as e:
            message, info = _get_db_error(e)
            raise logic.MoyaException("db.integrity-error", message, info=info)
        except OperationalError as e:
            message, info = _get_db_error(e)
            raise logic.MoyaException("db.operational-error", message, info=info)
        except StatementError as e:
            message, info = _get_db_error(e)
            raise logic.MoyaException("db.statement-error", message, info=info)
        except DatabaseError as e:
            message, info = _get_db_error(e)
            raise logic.MoyaException("db.error", message, info=info)
        except Exception as e:
            raise
        else:
            return ret
    return deco


class _SessionContextManager(object):
    def __init__(self, session, element):
        self._session = session
        self._element = element

    def __enter__(self):
        self._session.enter_transaction()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._session.exit_transaction(element=self._element,
                                       exc_type=exc_type,
                                       exc_val=exc_val)


class DBSession(object):
    def __init__(self, session_factory, engine=None):
        self.session_factory = scoped_session(session_factory)
        self.engine = engine
        self._session = None
        self._transaction_level = 0

    @property
    def session(self):
        if self._session is None:
            self._session = self.session_factory()
        return self._session

    def __moyacontext__(self, context):
        return self._session

    def manage(self, element):
        self.session
        return _SessionContextManager(self, element)

    def rollback(self):
        self.session.rollback()

    def __repr__(self):
        if self.session is not None:
            return '<dbsession %s>' % self.engine
        return '<dbsession>'

    def enter_transaction(self):
        self._transaction_level += 1

    @wrap_db_errors
    def exit_transaction(self, element=None, exc_type=None, exc_val=None):
        self._transaction_level -= 1
        if exc_type is None:
            try:
                self.session.commit()
            except:
                self.session.rollback()
                raise
        else:
            self.session.rollback()

    def __getattr__(self, key):
        return getattr(self.session, key)


def add_engine(archive, name, section):
    engine_name = section['engine']
    echo = attr_bool(section.get('echo', 'n'))
    default = attr_bool(section.get('default', 'n'))

    connect_args = {}
    if engine_name.startswith('sqlite:'):
        connect_args['check_same_thread'] = False

    sqla_engine = create_engine(engine_name,
                                echo=echo,
                                pool_recycle=3600,
                                connect_args=connect_args)

    # if engine_name.startswith('sqlite:'):
    #     @event.listens_for(sqla_engine, "connect")
    #     def do_connect(dbapi_connection, connection_record):
    #         # disable pysqlite's emitting of the BEGIN statement entirely.
    #         # also stops it from emitting COMMIT before any DDL.
    #         dbapi_connection.isolation_level = None

    #     @event.listens_for(sqla_engine, "begin")
    #     def do_begin(conn):
    #         # emit our own BEGIN
    #         conn.execute("BEGIN EXCLUSIVE")

    engine = DBEngine(name, engine_name, sqla_engine, default)

    if default or not archive.database_engines:
        archive.default_db_engine = name
    archive.database_engines[name] = engine
    startup_log.debug('%r created', engine)


def get_session_map(archive):
    """Get a dictionary that maps db names on to session objects"""
    session_map = {db: engine.get_session() for db, engine in archive.database_engines.items()}
    if archive.default_db_engine is not None:
        session_map['_default'] = session_map[archive.default_db_engine]
    return session_map


def commit_sessions(context):
    count = 0
    for dbsession in context['._dbsessions'].values():
        if dbsession.session:
            try:
                dbsession.session.commit()
            except:
                db_log.exception('error committing session')
            else:
                count += 1
    return count


def rollback_sessions(context):
    count = 0
    for dbsession in context['._dbsessions'].values():
        if dbsession.session:
            try:
                dbsession.session.rollback()
            except:
                db_log.exeption('error rolling back session')
            else:
                count += 1
    return count


def sync_all(archive, console):
    if validate_all(archive, console) != 0:
        return -1
    engines = archive.database_engines
    for engine in itervalues(engines):
        if engine.default:
            default_engine = engine
            break
    else:
        default_engine = None

    apps = archive.apps.values()

    try:
        with console.progress('syncing', num_steps=len(apps), width=24) as progress:
            progress.update(None, 'building models...')
            for app in apps:
                for model in app.lib.get_elements_by_type((namespaces.db, "model")):
                    model._build_model(app)

            synced = []
            for app in apps:
                progress.update(None, 'syncing {!r}'.format(app))
                for model in app.lib.get_elements_by_type((namespaces.db, "model")):

                    engine_name = model.dbname
                    if engine_name is None:
                        engine = default_engine
                    else:
                        engine = engines[engine_name]
                    model.create_all(archive, engine, app)
                progress.step()
                synced.append(app)

            progress.update(None, 'sync complete')
    finally:
        for app in synced:
            console(text_type(app), bold=True, fg="magenta")(" synced", bold=True, fg="black").nl()

    return 0


def validate_all(archive, console=None):
    """Validates models and returns the number of fails"""

    if not archive.database_engines:
        return 0

    from .tags.db import DBModel
    fails = DBModel.validate_all(archive)

    if console is None:
        return not len(fails)

    for model, app, element, error in fails:
        if element:
            console.document_error(text_type(error),
                                   element._location,
                                   element._code,
                                   element.source_line,
                                   None)
        else:
            console.error(text_type(error))
        if hasattr(error, 'diagnosis'):
            console.table([(error.diagnosis,)])

    return len(fails)



from __future__ import unicode_literals

from ..elements.registry import Meta
from ..elements.elementbase import ElementBase, Attribute, LogicElement, MoyaAttributeError
from ..elements.elementproxy import DataElementProxy
from ..logic import DeferNodeContents
from ..tags.context import _LazyCallable, DataSetter
from ..context.missing import is_missing
from ..compat import text_type, py2bytes
from .. import errors


class TagBase(LogicElement):

    def __init__(self, *args, **kwargs):
        self.init_args = (args, kwargs)
        super(TagBase, self).__init__(*args, **kwargs)

    def get_tag_app(self, context):
        app = None
        if self.has_parameter('from'):
            app = self.get_parameter(context, 'from') or None
        if app is None:
            try:
                app = self.archive.get_app_from_lib(self._lib_long_name)
            except Exception as e:
                self.throw("tag.ambiguous-app",
                           text_type(e),
                           diagnosis="You may need to supply the 'from' attribute")
        return app

    def logic(self, context):
        dst, lazy = self.get_parameters(context, 'dst', 'lazy')
        params = self.get_all_parameters(context)

        app = self.get_tag_app(context)

        if lazy and dst:
            element_callable = self.archive.get_callable_from_element(self.element, app=app)
            lazy_callable = _LazyCallable(context, element_callable, (), params)
            context.set_lazy(dst, lazy_callable)
        else:
            macro_call = self.push_call(context,
                                        {},
                                        app=app,
                                        yield_element=self,
                                        yield_frame=context.current_frame)
            try:
                macro_call.update(params)
                yield DeferNodeContents(self.element)
            finally:
                call = self.pop_call(context)
                if '_return' in call:
                    value = _return = call['_return']
                    if hasattr(_return, 'get_return_value'):
                        value = _return.get_return_value()
                else:
                    value = None
                if dst is None:
                    getattr(context.obj, 'append', lambda a: None)(value)
                else:
                    context[dst] = value


class Yield(LogicElement):
    """Yield from a tag"""

    class Help:
        synopsis = "yield to code block"

    scope = Attribute("scope to use", type="expression", default=None)

    def logic(self, context):
        scope = self.scope(context)
        if scope is None:
            scope = {}
        scope.update(self.get_let_map(context))

        call = context['.call']

        if is_missing(call) or not call.yield_frame:
            self.throw("yield.cant-yield",
                       "Can't yield from here")

        if scope:
            context.push_frame(call.yield_frame)
            try:
                with context.data_scope(scope):
                    yield DeferNodeContents(call.yield_element)
            finally:
                context.pop_frame()

        else:
            context.push_frame(call.yield_frame)
            try:
                yield DeferNodeContents(call.yield_element)
            finally:
                context.pop_frame()


class GetTagText(DataSetter):
    """
    Get the XML text associated with the parent tag.

    """

    class Help:
        synopsis = "get text from the parent tag"

    sub = Attribute("Substitute the text?", type="boolean", default=False)
    strip = Attribute("Strip text?", type="boolean", default=True)

    def get_value(self, context):
        call = context['.call']

        try:
            element = call.yield_element
        except AttributeError:
            self.throw('get-tag-text.error',
                       "Unable to retrieve tag text here")

        if element is None:
            self.throw('get-tag-text.error',
                       "Unable to detect parent tag")
        text = element.text
        if self.sub(context):
            context.push_frame(call.yield_frame)
            try:
                text = context.sub(text)
            finally:
                context.pop_frame()
        if self.strip(context):
            text = text.strip()
        return text


class Tag(ElementBase):
    """
    Define a custom tag. A custom tag is in a callable tag that works like builtin logic tags. Here's an example of defining a custom tag:

    [code xml]
    <tag name="getstock">
        <doc>Get count of available stock</doc>
        <signature>
            <attribute name="product" />
        </signature>
        <db:get model="#Product.name == product" dst="product"/>
        <return value="product.count"/>
    </tag>
    [/code]

    You may define a custom tag anywhere in a library. Once defined, the tag is available to any code that uses the appropriate namespace defined in [c]lib.ini[/c]. For example, if the above tag is defined in a library with an xml namespace of [c]http://moyaproject.com/sushifinder[/c], it could be invoked with the following file:

    [code xml]
    <moya xmlns:sushifinder="http://moyaproject.com/sushifinder">
        <macro name="main">
            <sushifinder:getstock product="tuna-roll" dst="count" />
            <echo>We have ${count} tuna rolls in stock.</echo>
        </macro>
    </moya>
    [/code]

    Custom tags are preferred over macros when exposing functionality to other libraries.

    """

    class Help:
        synopsis = "define a custom tag"

    ns = Attribute("XML Namespace", required=False, default=None)
    name = Attribute("Tag name", required=True)
    synopsis = Attribute("Short description of tag")
    undocumented = Attribute("Set to yes to disabled documentation for this tag", type="boolean", default=False)

    _tag_base = TagBase

    def finalize(self, context):
        params = self.get_parameters(context)
        attributes = {}
        try:
            for signature in self.children("signature"):
                for attribute_tag in signature.children("attribute"):
                    param_map = attribute_tag.get_all_parameters(context)
                    attribute = Attribute(attribute_tag.doc,
                                          map_to=param_map.get('name'),
                                          evaldefault=True,
                                          **param_map)
                    attributes[attribute.name] = attribute
        except MoyaAttributeError as e:
            raise errors.ElementError(text_type(e),
                                      element=self)
        if 'dst' not in attributes:
            attributes['dst'] = Attribute("Destination", name="dst", type="reference", map_to="dst", default=None)
        if 'lazy' not in attributes:
            attributes['lazy'] = Attribute("Enable lazy evaluation", name="lazy", type="boolean", map_to="lazy", default=False)
        attributes['from'] = Attribute("Application", name="from", type="application", map_to="from", default=None)
        doc = None
        for doc_tag in self.children("doc"):
            doc = doc_tag.text.strip()
        meta = Meta()
        meta.is_call = True

        cls_dict = dict(__doc__=text_type(doc or ''), Meta=meta)
        if self.source_line:
            definition = "%s (line %s)" % (self._location, self.source_line)
        else:
            definition = self._location

        class Help(object):
            synopsis = params.synopsis
            undocumented = params.undocumented

        ns = params.ns or self.lib.namespace
        if not ns:
            _msg = 'could not detect namespace for custom tag "{}" -- please specify the namespace with the "ns" attribute or in lib.ini'
            raise errors.ElementError(_msg.format(params.name),
                                      element=self)

        cls_dict.update({'Help': Help,
                         'xmlns': ns,
                         '_registry': self.archive.registry,
                         'element': self,
                         'libname': None,
                         '_definition': definition,
                         '_location': self._location,
                         'source_line': self.source_line,
                         '_code': self._code,
                         '_lib_long_name': context.get('._lib_long_name', None)})

        cls_dict.update({'_attribute_' + k: v for k, v in attributes.items()})

        tag_class = type(py2bytes(params.name),
                         (self._tag_base,),
                         cls_dict)

        tag_class._definition = definition


class DataTagBase(LogicElement):

    def __init__(self, *args, **kwargs):
        self.init_args = (args, kwargs)
        super(DataTagBase, self).__init__(*args, **kwargs)

    def get_proxy(self, context, app):
        return DataElementProxy(context,
                                app,
                                self,
                                self.get_all_data_parameters(context))

    def finalize(self, context):
        self.archive.add_data_tag(self._element_type, self)


class DataTag(ElementBase):
    """Define a data tag. See [doc customtags]."""

    ns = Attribute("XML Namespace", required=False, default=None)
    name = Attribute("Tag name", required=True)
    synopsis = Attribute("Short description of the data tag")
    undocumented = Attribute("Set to yes to disabled documentation for this tag", type="boolean", default=False)

    _tag_base = DataTagBase

    class Help:
        synopsis = "define a data tag"

    def finalize(self, context):
        params = self.get_parameters(context)
        attributes = {}
        for signature in self.children("signature"):
            for attribute_tag in signature.children("attribute"):
                param_map = attribute_tag.get_all_parameters(context)
                attribute = Attribute(attribute_tag.doc,
                                      map_to=param_map.get('name'),
                                      evaldefault=True,
                                      **param_map)
                attributes[attribute.name] = attribute
        doc = None
        for doc_tag in self.children("doc"):
            doc = doc_tag.text.strip()
        meta = Meta()
        meta.logic_skip = True
        meta.is_call = False

        class Help(object):
            synopsis = params.synopsis
            undocumented = params.undocumented

        cls_dict = dict(__doc__=text_type(doc or ''), Meta=meta, Help=Help)
        if self.source_line:
            definition = "%s (line %s)" % (self._location, self.source_line)
        else:
            definition = self._location

        cls_dict['_definition'] = definition
        cls_dict['xmlns'] = params.ns or self.lib.namespace
        cls_dict.update({'_attribute_' + k: v for k, v in attributes.items()})
        cls_dict['_registry'] = self.archive.registry

        tag_class = type(py2bytes(params.name),
                         (self._tag_base,),
                         cls_dict)
        tag_class.element = self
        tag_class._definition = definition



"""

Validate credit card numbers

http://en.wikipedia.org/wiki/Luhn_algorithm

"""


from __future__ import unicode_literals
from __future__ import print_function
from .compat import text_type


def checksum(card_number):
    """Calculate the luhn checksum"""
    def digits_of(n):
        return [int(d) for d in text_type(n)]
    digits = digits_of(card_number)
    odd_digits = digits[-1::-2]
    even_digits = digits[-2::-2]
    checksum = sum(odd_digits)
    for d in even_digits:
        checksum += sum(digits_of(d * 2))
    return checksum % 10


def validate(card_number):
    """Validate a card number"""
    card_number = ''.join(c for c in text_type(card_number) if c.isdigit())
    return checksum(card_number) == 0


if __name__ == "__main__":
    print(validate("4111 1111 1111 1111"))  # True
    print(validate("5500 0000 0000 0004"))  # True
    print(validate("6011 0000 0000 0004"))  # True
    print(validate("3088 0000 0000 0009"))  # True
    print(validate("5500 0000 0000 0005"))  # False



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import

from .containers import OrderedDict
from . import iniparse
from .compat import text_type, string_types, PY2, implements_to_string
from . import errors
from .tools import textual_list

from fs.path import normpath, relpath

import io
import os


def read_settings(fs, path):
    with fs.safeopen(path, 'rb') as settings_file:
        cfg = iniparse.parse(settings_file)
    return cfg


@implements_to_string
class SettingsKeyError(KeyError):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class SettingsContainer(OrderedDict):

    @classmethod
    def apply_master(self, master, settings):
        for section_name, section in master.items():
            if section_name == 'service':
                continue
            if section_name in settings:
                settings[section_name].update(section)
            else:
                settings[section_name] = settings

    @classmethod
    def read(cls, fs, path, master=None):
        visited = []

        if not isinstance(path, string_types):
            for p in path:
                if fs.isfile(p):
                    path = p
                    break
            else:
                raise errors.SettingsError('''settings file not found (looked for {} in {})'''.format(textual_list(path, join_word='and'), fs))

        settings_stack = []
        while 1:
            path = relpath(normpath(path))
            if path in visited:
                raise errors.SettingsError('''recursive extends detected, "{}" has already been extended'''.format(path))
            with fs.open(path, 'rt') as settings_file:
                s = iniparse.parse(settings_file,
                                   SettingsContainer(),
                                   section_class=SettingsSectionContainer)
            visited.append(path)
            settings_stack.append(s)
            if "extends" in s['']:
                path = s['']['extends']
            else:
                break
        settings_stack = settings_stack[::-1]
        settings = settings_stack[0]
        s = cls.__class__(settings_stack[0])
        for s in settings_stack[1:]:
            for section_name, section in s.items():
                if section_name in settings:
                    settings[section_name].update(section)
                else:
                    settings[section_name] = section

        if master is not None:
            cls.apply_master(master, settings)
        return settings

    @classmethod
    def read_os(self, path):
        visited = []
        settings_stack = []
        while 1:
            path = os.path.abspath(os.path.normpath(path))
            if path in visited:
                raise errors.SettingsError('''recursive extends detected, "{}" has already been extended'''.format(path))
            with io.open(path, 'rt') as settings_file:
                s = iniparse.parse(settings_file,
                                   SettingsContainer(),
                                   section_class=SettingsSectionContainer)
            visited.append(path)
            settings_stack.append(s)
            if "extends" in s['']:
                path = s['']['extends']
            else:
                break
        settings_stack = settings_stack[::-1]
        settings = settings_stack[0]
        s = self.__class__(settings_stack[0])
        for s in settings_stack[1:]:
            for section_name, section in s.items():
                if section_name in settings:
                    settings[section_name].update(section)
                else:
                    settings[section_name] = section
        return settings

    @classmethod
    def read_from_file(self, settings_file):
        """Reads settings, but doesn't do any extends processing"""
        settings = iniparse.parse(settings_file,
                                  SettingsContainer(),
                                  section_class=SettingsSectionContainer)
        return settings

    @classmethod
    def from_dict(self, d):
        return SettingsSectionContainer((k, SettingContainer(v)) for k, v in d.items())

    @classmethod
    def create(cls, **kwargs):
        return cls.from_dict(kwargs)

    def export(self, output_file, comments=None):
        """Write the settings to an open file"""
        ini = iniparse.write(self, comments=comments)
        output_file.write(ini)

    def copy(self):
        return SettingsContainer(self)

    def __getitem__(self, key):
        try:
            return super(SettingsContainer, self).__getitem__(key)
        except KeyError:
            return EmptySettings()

    def get(self, section_name, key, default=Ellipsis):
        if section_name not in self:
            if default is Ellipsis:
                raise SettingsKeyError("Required section '%s' not found in settings" % section_name)
            else:
                return default
        section = self[section_name]
        if key not in section:
            if default is Ellipsis:
                raise SettingsKeyError("Key '%s' not found in section '%s'" % (key, section_name))
            else:
                return default
        return section[key]

    def set(self, section_name, key, value):
        if section_name not in self:
            self[section_name] = SettingsSectionContainer()
        self[section_name][key] = value

    def get_bool(self, section_name, key, default=False):
        value = self.get(section_name, key, "yes" if default else "no")
        return value.strip().lower() in ("yes", "true")

    def get_list(self, section_name, key, default=""):
        value = self.get(section_name, key, default=default)
        return [line.strip() for line in value.splitlines() if line.strip()]

    def get_int(self, section_name, key, default=None):
        value_text = self.get(section_name, key, None)
        if value_text is None or not value_text.strip():
            return None
        try:
            value = int(value_text)
        except:
            raise SettingsKeyError("Key [{}]/{} should be empty or a value integer (not '{}')".format(section_name, key, value_text))
        else:
            return value

    def __moyaconsole__(self, console):
        from console import Cell
        table = [(Cell("key", bold=True), Cell("value", bold=True))]
        table += sorted(self.items())
        console.table(table)


class SettingsSectionContainer(OrderedDict):

    def get_int(self, key, default=None):
        if key not in self:
            return default
        value = int(self[key])
        return value

    def get_bool(self, key, default=False):
        value = self.get(key, "yes" if default else "no")
        return value.strip().lower() in ("yes", "true")

    def get_list(self, key, default=""):
        value = self.get(key, default)
        return [line.strip() for line in value.splitlines() if line.strip()]

    def __moyaconsole__(self, console):
        from console import Cell
        table = [(Cell("key", bold=True), Cell("value", bold=True))]
        table += sorted(self.items())
        console.table(table)


class EmptySettings(object):
    def __getitem__(self, key):
        if key == "list":
            return []
        if key == "bool":
            return False
        if key == "int":
            return 0
        return ''

    def get_int(self, key, default=None):
        return default

    def get_bool(self, key, default=False):
        return default

    def get_list(self, key, default=""):
        return default

    def get(self, key, default=None):
        return default

    def __unicode__(self):
        return ''

    def __iter__(self):
        return iter([])

    def items(self):
        return []

    def __moyaconsole__(self, console):
        from console import Cell
        table = [(Cell("key", bold=True), Cell("value", bold=True))]
        console.table(table)


implements_to_string
class SettingContainer(text_type):
    def __init__(self, setting_text):
        if PY2:
            super(SettingContainer, self).__init__(setting_text)
        else:
            super().__init__()
        #self.raw = setting_text
        self.setting_text = setting_text.strip()
        self.lines = [line.strip() for line in setting_text.splitlines()] or []
        self.first = self.lines[0] if self.lines else ''
        self.bool = self.setting_text.lower() in ('yes', 'true')
        try:
            self.int = int(self.setting_text)
        except ValueError:
            self.int = None
        try:
            self.float = float(self.setting_text)
        except ValueError:
            self.float = None

    def __str__(self):
        return self.setting_text

    def __getitem__(self, index):
        if isinstance(index, string_types):
            if index == "list":
                return self.lines
            elif index == "bool":
                return self.bool
            elif index == "int":
                return self.int
            elif index == "float":
                return self.float
        return self.lines[index]

    def __eq__(self, other):
        return self.first == other

    def __ne__(self, other):
        return self.first != other

    if not PY2:
        def __hash__(self):
            return super().__hash__()


if __name__ == "__main__":

    settings = SettingsContainer()
    print(settings['nothere'])

    s = SettingContainer("foo\nbar")
    print(s == "foo")
    print(s['list'])



from __future__ import unicode_literals
from __future__ import print_function

from .compat import text_type, implements_to_string
from .tools import textual_list


class ArchiveError(Exception):
    """Error occurred with an archive operation"""


class LogicError(Exception):
    def __init__(self, original, trace):
        self.original = original
        self.moya_trace = trace
        super(LogicError, self).__init__(trace.msg)

    @property
    def diagnosis(self):
        return self.original.diagnosis

    def __moyaconsole__(self, console):
        self.moya_trace.__moyaconsole__(console)

    def __unicode__(self):
        return self.moya_trace.msg

    def __repr__(self):
        return "<LogicError {}>".format(self.moya_trace.msg)


class MoyaError(Exception):
    """Base exception for Moya related errors"""

    # The message to use if no error is supplied
    default_message = ""
    # The message to user if an error is supplied
    message = "{error}"

    def _escape_format(cls, text):
        return text.replace('{', '{{').replace('}', '}}')

    def __init__(self, error=None, **kwargs):
        fmt_args = kwargs
        fmt_args['error'] = error
        msg = self.format_error(error, fmt_args)
        if 'diagnosis' in kwargs:
            self.diagnosis = kwargs['diagnosis']
        super(MoyaError, self).__init__(msg)

    def format_error(self, error, fmt_args):
        if fmt_args:
            fmt_args = {k: self._escape_format(text_type(v)) for k, v in fmt_args.items()}
        if error is None:
            msg = self.default_message.format(**fmt_args)
        else:
            # print error, repr(fmt_args)
            # error = self._escape_format(error)
            fmt_args['error'] = error.format(**fmt_args)
            msg = self.message.format(**fmt_args)
        return msg

    def __moyaconsole__(self, console):
        console.text(text_type(self))

    # def __unicode__(self):
    #     if self.fmt is None:
    #         return super(MoyaError, self).__unicode__()
    #     else:
    #         return self.fmt.format(**self.get_fmt_dict())

    # def __str__(self):
    #     if self.fmt is None:
    #         return super(MoyaError, self).__str__()
    #     else:
    #         return self.fmt.format(**self.get_fmt_dict()).encode('ascii', 'replace')


class ParseError(MoyaError):
    def __init__(self, message, path="?", position=None, code=None):
        super(ParseError, self).__init__(message)
        self.path = path
        line, col = position
        self.position = position
        self.line = line
        self.col = col
        self.code = code

    def render(self, colour=False):
        line, col = self.position
        lines = self.code.replace('\t', ' ' * 4).splitlines()
        start = max(0, line - 3)
        end = min(len(lines), line + 2)
        showlines = lines[start:end]

        linenos = [str(n + 1) for n in range(start, end)]
        maxline = max(len(l) for l in linenos)

        errorlines = []
        highlight_line = str(line)
        for lineno, line in zip(linenos, showlines):
            if lineno == highlight_line:
                fmt = "*%s %s"
            else:
                fmt = " %s %s"
            errorlines.append(fmt % (lineno.ljust(maxline), line))

        print('\n'.join(errorlines))


@implements_to_string
class DocumentError(MoyaError):
    """Raised when there is an error constructing the document"""
    def __init__(self, element, msg=''):
        super(DocumentError, self).__init__()
        self.element = element
        self.msg = msg

    def __repr__(self):
        return '%s in %s' % (self.msg, self.element)


class AttributeError(MoyaError):
    """An attribute related parse error"""
    hide_py_traceback = True
    error_type = "Attribute error"


class BadValueError(MoyaError):
    hide_py_traceback = True
    error_type = "Invalid attribute error"


@implements_to_string
class ElementError(MoyaError):
    error_type = "Element error"
    hide_py_traceback = True

    def __init__(self, msg=None, element=None, diagnosis=None):
        self.msg = msg
        self.element = element
        self.diagnosis = diagnosis
        super(ElementError, self).__init__(msg)

    @property
    def source_line(self):
        if self.element:
            return getattr(self.element, 'source_line', None)
        return None

    def get_message(self):
        if self.element is None:
            return self.msg
        #path = self.element._document.path
        #line = self.element.source_line or '?'
        return 'in {}, {}'.format(self.element, self.msg)
        # return 'Document "%s", line %s, in <%s>: %s' % (path,
        #                                                 line,
        #                                                 self.element._tag_name,
        #                                                 self.msg)

    def __str__(self):
        return text_type(self.get_message())


class ContentError(ElementError):
    error_type = "Content error"


class ElementNotFoundError(MoyaError):
    default_message = "element '{elementref}' was not found in the project"
    hide_py_traceback = True
    error_type = "Element not found error"

    def __init__(self, elementref, app=None, lib=None, msg=None):
        if isinstance(elementref, tuple):
            xmlns, ref = elementref
            elementref = "{{" + xmlns + "}}" + ref
        self.elementref = elementref
        self.app = app
        self.lib = lib
        if msg is None:
            if app or lib:
                msg = "element '{elementref}' not found in {obj}".format(elementref=self.elementref,
                                                                         obj=self.app or self.lib)
            else:
                msg = "element '{elementref}' was not found".format(elementref=self.elementref)
        else:
            msg = msg.replace('{', '{{').replace('}', '}}')
        super(ElementNotFoundError, self).__init__(msg, elementref=elementref)

    # def get_message(self):
    #     if not (self.app or self.lib):
    #         return super(ElementNotFoundError, self).get_message()
    #     return "element '{elementref}' not found in {obj}".format(elementref=self.elementref,
    #                                                               obj=self.app or self.lib)


class UnknownLibraryError(MoyaError):
    default_message = "library '{lib}' must be imported before it can be installed"
    hide_py_traceback = True
    error_type = "Library not imported error"

    def __init__(self, lib):
        self.lib = lib
        super(UnknownLibraryError, self).__init__(lib=lib)


class UnknownElementError(MoyaError):
    default_message = "element {{{xmlns}}}{element} is not recognized"

    def __init__(self, xmlns, element, source_line=None):
        self.xmlns = xmlns
        self.element = element
        self.source_line = source_line
        super(UnknownElementError, self).__init__(xmlns=xmlns, element=element)


class AmbiguousFilterError(MoyaError):
    default_message = "filter is ambigious"


class UnknownFilterError(MoyaError):
    default_message = "no such filter"


class AttributeTypeError(MoyaError):
    """An error caused by an attribute containing the wrong type of data"""
    def __init__(self, element, name, value, type_name):
        self.element = element
        self.name = name
        self.type_name = type_name
        self.value = value
        msg = "%r attribute should be a valid %s (not %r)" % (name,
                                                              type_name,
                                                              value)
        super(AttributeTypeError, self).__init__(msg)


class ContextError(MoyaError):
    pass


class LibraryLoadError(MoyaError):
    """Raised when a lib could not be read"""
    hide_py_traceback = True
    default_message = "Unable to load library '{lib}'"
    message = "Unable to load library '{lib}' - {error}"

    def __init__(self, error, lib=None, py_exception=None, **kwargs):
        long_name = getattr(lib, 'long_name', None)
        if long_name is None:
            lib = "<unknown>"
        else:
            lib = long_name

        self.lib = lib
        self.py_exception = py_exception
        super(LibraryLoadError, self).__init__(error, lib=lib, **kwargs)


class StartupFailedError(MoyaError):
    pass


class SettingsError(StartupFailedError):
    pass


class LoggingSettingsError(StartupFailedError):
    pass


class AppError(MoyaError):
    hide_py_traceback = True
    error_type = "Application Error"


class AmbiguousAppError(AppError):
    default_message = "More than one app installed for lib '{lib_name}', choices are {apps}"

    def __init__(self, lib_name, apps):
        self.lib_name = lib_name
        self.apps = apps
        super(AmbiguousAppError, self).__init__(lib_name=lib_name, apps=textual_list(apps))


class AppRequiredError(AppError):
    default_message = "No application installed for lib '{lib}'"

    def __init__(self, lib):
        super(AppRequiredError, self).__init__(lib=lib)


class AppMissingError(AppError):
    default_message = "A value for application is required"


class UnknownAppError(AppError):
    default_message = "No app in the project referenced by '{app}'"



from __future__ import unicode_literals
from __future__ import print_function

from ..elements.elementbase import Attribute, LogicElement
from ..tags.context import ContextElementBase, DataSetter
from ..render import HTML, is_renderable, Unsafe, render_object, is_safe
from ..context.dataindex import makeindex
from .. import logic
from .. import errors
from ..html import escape
from ..console import Console
from ..content import Content, Section, IncludePath
from ..tools import url_join
from ..context.missing import is_missing
from ..markup import Markup
from ..template import Template as MoyaTemplate
from ..compat import string_types, text_type


class Renderable(object):
    moya_render_targets = ["html"]


class TemplateRenderable(Renderable):
    moya_safe = True

    def __init__(self, template, td):
        self.template = template
        self.td = td
        self.children = []

    def moya_render(self, archive, context, target, options):
        engine = archive.get_template_engine("moya")
        template = self.template
        if isinstance(template, string_types):
            html = engine.render(template, self.td, base_context=context)
        else:
            html = engine.render_template(template, self.td, base_context=context)
        return HTML(html)

    def add_renderable(self, renderable):
        self.children.append(renderable)


class ContentElementMixin(object):

    def resolve_template(self, app, template, element=None):
        if element is None:
            element = self
        if template is None:
            return None
        template = app.resolve_template(template)
        if not self.archive.get_template_engine().exists(template):
            raise errors.ContentError("missing template '{}'".format(template),
                                      element=element,
                                      diagnosis="You can check what templates are installed with **moya fs templates --tree**.")
        return template

    def push_content_frame(self, context, content):
        content_stack = context.set_new_call(".contentstack", list)
        content_stack.append(content)
        context['.content'] = content
        template_data_index = makeindex('.contentstack', len(content_stack) - 1, 'td')
        context.push_scope(template_data_index)

    def pop_content_frame(self, context):
        context.pop_scope()
        stack = context['.contentstack']
        value = stack.pop()
        if stack:
            context['.content'] = stack[-1]
        else:
            del context['.content']
        return value

    def get_content(self, context):
        content_stack = context['.contentstack']

        if not content_stack:
            raise logic.FatalMoyaException("content.content-not-found",
                                           "Content not found (did you forget the <content> tag)?")

        return content_stack[-1]

    def generate_content(self, context, element_ref, app, td):
        app, element = self.get_element(element_ref, app or None)

        merge_content = []
        for content_app, content_element in element.get_extends_chain(context, app=app):
            template = self.resolve_template(content_app, content_element.template, content_element) if content_app else content_element.template
            content = Content(content_app, template, td=td)
            merge_content.append(content)

            if content_element.has_children:
                self.push_content_frame(context, content)
                try:
                    yield logic.DeferNodeContents(content_element)
                finally:
                    self.pop_content_frame(context)

        content = merge_content[0]
        for extended_content in merge_content[1:]:
            content.merge(extended_content)
        if content.template is None:
            content.template = app.default_template
        if not content.template:
            raise errors.ElementError('content has no template',
                                      element=self,
                                      diagnosis="You can specify a template on the &lt;content&gt; definition")
        context['_content'] = content


class ContentElement(ContextElementBase):
    """
    Begin a [link content]content[/link] definition.

    Content is a high level description of a page.

    """

    class Help:
        synopsis = "define content"
        example = """
        <content libname="content.crew.manifest">
            <title>Crew Manifest</title>
            <section name="content">
                <for src="crew" dst="character">
                    <moya:crew character="character"/>
                </form>
            </section>
            <section name="footer">
                <markdown>Brought to you by **Moya**.</markdown>
            </section>
        </content>

        """

    template = Attribute("Template name", type="template", required=False, default=None)
    extends = Attribute("Extend content element", type="elementref")

    preserve_attributes = ['template', 'extends', '_merge']

    class Meta:
        tag_name = "content"
        translate = False

    def get_extends_chain(self, context, app=None):
        element_refs = set()
        app = app or context.get('.app', None)
        node = self
        nodes = [(app, node)]
        extends_ref = node.extends

        while extends_ref:
            element_refs.add(extends_ref)
            node_app, node = self.document.get_app_element(extends_ref, app=app)
            app = node_app or app
            if node is None:
                break
            nodes.append((app, node))
            extends_ref = node.extends
            if extends_ref in element_refs:
                raise errors.ContentError("element '{}' has already been extended".format(extends_ref),
                                          element=self,
                                          diagnosis="Check the 'extends' attribute in your content tags.")

        base_content = context.get('.sys.site.base_content')
        if base_content:
            app, node = self.document.get_element(base_content,
                                                  lib=node.lib)
            nodes.append((app, node))

        chain = nodes[::-1]
        # for app, node in chain:
        #     print (repr(app), repr(node), node.template)
        return chain

    def post_build(self, context):
        self.template = self.template(context)
        self.extends = self.extends(context)


class SectionElement(LogicElement, ContentElementMixin):
    """Defines a section container for content. A [i]section[/i] is a top level container for content, and is used to break up the content in to function groups, which may be rendered independently. For example, here is a content definition with two sections; 'body' and 'sidebar':

    [code xml]
    <content libname="content.front" template="front.html">
        <section name="body">
            <!-- main body content goes here -->
        </section>
        <section name="sidebar">
            <!-- sidebar content here -->
        </section>
    </content>
    [/code]

    The template for the above content would render the sections with the [link templates#render]{% render %}[/link] template tag, For example:

    [code moyatemplate]
    <html>
        <body>
            <div id="sidebar>
            {% render sections.sidebar %}
            </div>
            <h1>Content example</h1>
            {% render sections.body %}
        </body>
    </html>
    [/code]


    """

    class Help:
        synopsis = "create a content section"

    name = Attribute("The name of the section", required=True)
    template = Attribute("Template", required=False, default=None)
    merge = Attribute("Merge method", default="inherit", choices=["inherit", "append", "prepend", "replace"])

    preserve_attributes = ['_merge']

    class Meta:
        tag_name = "section"
        #logic_skip = True
        #text_nodes = "text"

    def logic(self, context):
        name = self.name(context)
        name, template = self.get_parameters(context, 'name', 'template')
        content = self.get_content(context)
        app = self.get_app(context)
        content.new_section(name, app.resolve_template(template), merge=self._merge)

        with content.section(name):
            yield logic.DeferNodeContents(self)

    def post_build(self, context):
        self._merge = self.merge(context)


class Node(LogicElement, ContentElementMixin):
    """
    Create a template node in a content definition. A node is essentially a reference to a template with associated data. Here's an example of a content definition containing a template node:

    [code xml]
    <content libname="content.example" template="front.html">
        <section name="body">
            <node template="newscontainer.html" let:style="splash"/>
                <html:p>${news}</html:p>
            </node>
        </section>
    </content>
    [/code]

    Here's what newscontainer.html might look like. Note the use of [link templates#children]{% children %}[/link] which will render the content contained inside a node:

    [code moyatemplate]
    <div style="${style}">
        <h3>The Latest News</h3>
        {% children %}
    </div>
    [/code]
    """

    class Help:
        synopsis = """creates a template node in content"""

    template = Attribute("Template", type="template", required=False, default=None)
    withscope = Attribute("Is current context when rendering template", type="boolean", default=False, required=False)
    _from = Attribute("Application", type="application", required=False, default=None)
    ifexists = Attribute("Skip if the template can not be found", type="boolean", required=False, default=False)

    def logic(self, context):
        template, withscope, ifexists = self.get_parameters(context, 'template', 'withscope', 'ifexists')
        content = self.get_content(context)
        app = self.get_app(context)
        template = app.resolve_template(template)

        if ifexists:
            engine = self.archive.get_template_engine("moya")
            if not engine.exists(template):
                return

        if withscope:
            td = {}
            td.update(context.capture_scope())
        else:
            td = self.get_let_map(context)
        content.add_template("node", template, td, app=app)
        if self.has_children:
            with content.node():
                yield logic.DeferNodeContents(self)


class ScopeNode(Node):
    """
    Create a template node that uses the current scope as the template data.

    This node is identical to [tag]node[/tag], with the [c]withscope[/c] attribute set to [c]yes[/c].

    """

    class Help:
        synopsis = """creates a template node using the current scope"""
    withscope = Attribute("Is current context when rendering template", type="boolean", default=True, required=False)


class RenderBase(LogicElement):
    obj = Attribute("Object to render", type="index")

    class Help:
        undocumented = True

    class Meta:
        translate = True

    def render_children(self, context):
        for child in self.children(element_class="renderable"):
            pass

    def include_css(self, context, media, app, path):
        if isinstance(app, text_type):
            app = self.archive.get_app_from_lib(app)
        path = self.archive.get_media_url(context, app, media, path)
        content = context['.content']
        content.include('css', IncludePath('css', path, IncludeCSS.format))


class RenderContent(DataSetter, ContentElementMixin):
    """Render content"""

    class Help:
        synopsis = "render content"

    class Meta:
        is_call = True

    content = Attribute("Reference to renderable content", type="elementref", required=False, default=None)
    _from = Attribute("Application", type="application", required=False, default=None)
    withscope = Attribute("Use current scope?", default=False, type="boolean")
    template = Attribute("Template", required=False, default=None)

    def logic(self, context):
        (content,
         withscope,
         template,
         dst) = self.get_parameters(context,
                                    'content',
                                    'withscope',
                                    'template',
                                    'dst')
        app = self.get_app(context)
        template = app.resolve_template(template)
        if withscope:
            scope = context['.call']
        let = self.get_let_map(context)
        td = {}
        if self.has_children:
            call = self.push_funccall(context)
            try:
                yield logic.DeferNodeContents(self)
            finally:
                self.pop_funccall(context)

            args, kwargs = call.get_call_params()
            if withscope:
                new_kwargs = scope
                new_kwargs.update(kwargs)
                kwargs = new_kwargs
            td.update(kwargs)
        kwargs.update(let)

        for defer in self.generate_content(context, content, app, td=td):
            yield defer

        content_obj = context['_content']

        result = render_object(content_obj, self.archive, context, "html")
        self.set_context(context, dst, result)


class ServeContent(LogicElement, ContentElementMixin):
    """Render content and immediately serve it. Note that this tag will stop processing any more logic code."""

    class Help:
        synopsis = "render and serve content"
        example = """
        <serve-content content="#content.front" let:date=".now"/>

        """

    content = Attribute("Reference to renderable content", type="elementref", required=False, default=None)
    _from = Attribute("Application", type="application", required=False, default=None)
    withscope = Attribute("Use current scope?", default=False, type="boolean")
    template = Attribute("Template", required=False, default=None)

    class Meta:
        is_call = True

    def logic(self, context):
        content, withscope, template = self.get_parameters(context,
                                                           'content',
                                                           'withscope',
                                                           'template')
        app = self.get_app(context)
        template = app.resolve_template(template)
        if withscope:
            scope = context['.call']
        let = self.get_let_map(context)
        td = {}
        if self.has_children:
            call = self.push_funccall(context)
            try:
                yield logic.DeferNodeContents(self)
            finally:
                self.pop_funccall(context)

            args, kwargs = call.get_call_params()
            if withscope:
                new_kwargs = scope
                new_kwargs.update(kwargs)
                kwargs = new_kwargs
            td.update(kwargs)
        td.update(let)

        for defer in self.generate_content(context, content, app, td=td):
            yield defer
        context.copy('_content', '_return')

        raise logic.Unwind()


class Title(LogicElement):
    """Set the title for content. This tag simply sets a value called [c]title[/c] on the context, which can be rendered in a templates. Here's an example:

    [code xml]
    <content libname="content.front" template="front.html">
        <title>Welcome!<title>
    </content>
    [/code]

    A reference to [c]title[/c] would appear somewhere in the template associated with the content. For example:

    [code moyatemplate]
    <html>
       <head>
           <title>${title}</title>
        </head>
        <body>
            {% render sections.body %}
        </body>
    </html>
    [/code]
    """

    class Help:
        synopsis = "set content title"

    class Meta:
        translate = True

    def logic(self, context):
        context['title'] = context.sub(self.text)


class SimpleRenderable(Renderable):
    moya_safe = True

    class Help:
        undocumented = True

    def __init__(self, format_string, **kwargs):
        self.html = format_string.format(**kwargs)

    def __repr__(self):
        html = self.html
        if len(self.html) > 50:
            html = html[:50] + "[...]"
        return '<SimpleRenderable "%s">' % html

    def moya_render(self, archive, context, target, options):
        return HTML(self.html)


class MediaURL(DataSetter):
    """Get URL to media"""

    class Help:
        synopsis = "get a URL to media"

    path = Attribute("Path in media")
    media = Attribute("Media name", required=False, default='media')
    _from = Attribute("Application containing the media", type="application", default=None)
    dst = Attribute("Destination to store media URL", required=False, type="reference")

    def logic(self, context):
        params = self.get_parameters(context)
        app = self.get_app(context)
        media_path = self.archive.get_media_url(context, app, params.media)
        url = url_join(media_path, params.path)
        self.set_context(context, params.dst, url)


class IncludeCSS(LogicElement, ContentElementMixin):
    """
    Add a CSS path to be included in the content. The list of paths will be added to a value called [c]include[/c] when the template is rendered. Here's an example:

    [code xml]
    <content libname="content.front" template="front.html">
        <include-css path="css/news.css" />
    </content>
    [/code]

    The CSS paths can be rendered in a template as follows:

    [code moyatemplate]
    {% render include.css %}
    [/code]

    """

    class Help:
        synopsis = """include CSS with content"""

    type = Attribute("Type of link", required=False, default="css")
    media = Attribute("Media name", required=False, default='media')
    path = Attribute("Path to CSS", required=False, default=None)
    _from = Attribute("Application", type="application", default=None)
    url = Attribute("External URL", required=False, default=None)

    class Meta:
        one_of = [('path', 'url')]

    format = """<link href="{path}" rel="stylesheet" type="text/css">"""

    def logic(self, context):
        params = self.get_parameters(context)
        content = self.get_content(context)
        app = self.get_app(context)
        if params.url:
            path = params.url
        else:
            if params.path.startswith('/'):
                path = params.path
            else:
                media_path = self.archive.get_media_url(context, app, params.media)
                path = url_join(media_path, params.path)
        content.include(params.type, IncludePath(params.type, path, self.format))


class IncludeJS(IncludeCSS):
    """
    Like [tag]include-css[/tag], but inserts a link to a JS file.

    The JS files may be inserted in to the template as follows:

    [code moyatemplate]
    {% render include.js %}
    [/code]

    This is equivalent to the following:

    [code moyatemplate]
    {%- for include in include.js %}
    <script type="text/javascript" href="${include.path}"/>
    {%- endfor %}
    [/code]

    """

    class Help:
        synopsis = """include a JS file in content"""

    type = Attribute("Type of link", required=False, default="js")
    # media = Attribute("Media name", required=False, default='media')
    # path = Attribute("Path to JS", required=False, default=None)
    # _from = Attribute("Application", type="application", default=None)
    format = """<script src="{path}"></script>"""


class Render(DataSetter, ContentElementMixin):
    """
    Render a [i]renderable[/i] object.

    """

    class Help:
        synopsis = "render an object in content"
        example = """
        <render src="form" />
        """

    src = Attribute("Object to render", required=False, type="expression")
    dst = Attribute("Destination to store rendered content", required=False, type="reference")
    target = Attribute("Render target", required=False, default="html")

    def logic(self, context):
        content_container = context.get('.content', None)
        src, dst, target = self.get_parameters(context,
                                               'src',
                                               'dst',
                                               'target')

        if src is None:
            section = Section(None, td=self.get_let_map(context), name=self.libid)
            self.push_content_frame(context, section)
            try:
                yield logic.DeferNodeContents(self)
            finally:
                self.pop_content_frame(context)
            obj = section
        else:
            obj = src

        if not is_renderable(obj) and not is_safe(obj):
            obj = Unsafe(obj)
        if content_container is not None:
            content_container.add_renderable(self._tag_name, obj)
        else:
            rendered = render_object(obj, self.archive, context, target)
            self.set_context(context, dst, rendered)


class RenderAll(DataSetter, ContentElementMixin):
    """
    Render a sequence of renderable objects.
    """

    class Help:
        synopsis = "render a sequence of renderable objects"

    src = Attribute("Object to render", required=False, type="expression")
    dst = Attribute("Destination to store rendered content", required=False, type="reference")
    target = Attribute("Render target", required=False, default="html")

    def logic(self, context):
        content_container = context.get('.content', None)
        src, dst, target = self.get_parameters(context,
                                               'src',
                                               'dst',
                                               'target')

        try:
            obj_iter = iter(src)
        except:
            self.throw('render-all.not-a-sequence',
                       'src is not a sequence',
                       diagnosis="Moya was unable to iterate over {}".format(context.to_expr(src)))

        for obj in obj_iter:
            if not is_renderable(obj) and not is_safe(obj):
                obj = Unsafe(obj)
            if content_container is not None:
                content_container.add_renderable(self._tag_name, obj)
            else:
                rendered = render_object(obj, self.archive, context, target)
                self.set_context(context, dst, rendered)


class Template(RenderBase):

    markup = Attribute("Markup", required=False, default="html", choices=["bbcode", "html"])

    def finalize(self, context):
        self.template = MoyaTemplate(self.text, self._location)

    def logic(self, context):
        rendered_content = self.template.render(context)
        markup = Markup(rendered_content, self.markup(context))
        context['.content'].add_renderable(self._tag_name, markup)


class Raw(RenderBase):
    #name = Attribute("Template name of the text")

    def logic(self, context):
        context['.content'].add_renderable(self._tag_name, HTML(context.sub(self.text)))


class DefaultMarkup(RenderBase):
    """Use default markup if a value is None or missing"""

    class Help:
        synopsis = "use default markup for missing values"

    value = Attribute("Value", required=True, type="expression")
    default = Attribute("Default", required=False, default="&ndash;")

    def logic(self, context):
        value = self.value(context)
        if is_missing(value) or value is None:
            context['.content'].add_renderable(self._tag_name, HTML(self.default(context)))
        else:
            yield logic.DeferNodeContents(self)


class JS(RenderBase):
    """Insert Javascript content.

    Here's an example:

    [code xml]
    <!-- must be inside a content definition -->
    <js>alert("Ready for takeoff!);</js>
    [/code]

    This would render the following:

    [code moyatemplate]
    <script type="text/javascript">
        alert("Ready for takeoff!);
    </script>
    [/code]

    """

    class Help:
        synopsis = "insert Javascript content"

    class Meta:
        translate = False

    section = Attribute("Section to add script to", required=False, default="js")

    def logic(self, context):
        section = self.section(context)
        js = context.sub(self.text)
        html = """<script type="text/javascript">{}</script>\n""".format(js)
        context['.content'].get_section(section).add_renderable(self._tag_name, HTML(html))


class CSS(RenderBase):
    """
    This content tag creates a [c]<style>[/c] element in html, with the enclosed text.

    It is generally preferable to use [tag]include-css[/tag], but this tag can be useful to insert dynamically generated CSS.

    """
    section = Attribute("Section to add CSS tag to", required=False, default="css")

    class Help:
        synopsis = "add a CSS to content"
        example = """
        <css>
            .character.rygel
            {
                font-weight:bold
            }
        </css>
        """

    class Meta:
        translate = False

    def logic(self, context):
        section = self.section(context)
        css = context.sub(self.text)
        html = """<style type="text/css">%s</style>\n""" % css
        context['.content'].get_section(section).add_renderable(self._tag_name, HTML(html))


class Text(RenderBase):

    def logic(self, context):
        text = self.lib.translate(context, self.text)
        text = escape(context.sub(text))
        html = HTML(text or ' ')
        context['.content'].add_renderable(self._tag_name, html)


class ConsoleRender(RenderBase):
    """Render an object as terminal output. Useful as a debugging aid to quickly render an object."""
    obj = Attribute("Object to render", type="expression", required=True)

    def logic(self, context):
        obj = self.obj(context)
        c = Console(nocolors=False, text=False, width=120, html=True)
        text = c.obj(context, obj).get_text()
        html = '<div class="moya-console">{}</div>'.format(text)
        context['.content'].add_renderable(repr(obj), HTML(html))
        self.include_css(context, "media", 'moya.debug', 'css/debug.css')
